import copy
import pydoc

from plynx.db.node import Node
from plynx.base import hub
from plynx.utils.common import parse_search_string
from plynx.utils.hub_node_registry import registry
import plynx.node


def _enhance_list_item(raw_item):
    if raw_item['_type'] == 'Group':
        # TODO proper checking
        items = []
        for raw_subitem in raw_item['items']:
            items.append(_enhance_list_item(raw_subitem))
        raw_item['items'] = items
        return raw_item
    # check if the node is valid
    node = Node.from_dict(raw_item)
    registry.register_node(node)
    return node.to_dict()


def _recursive_filter(search_string, list_of_nodes):
    res = []
    for raw_node in list_of_nodes:
        if raw_node["_type"] == "Group":
            raw_group = copy.deepcopy(raw_node)
            raw_group['items'] = _recursive_filter(search_string, raw_group['items'])
            res.append(raw_group)
        elif len(search_string) == 0 or search_string.upper() in raw_node['title'].upper():
            res.append(raw_node)
    return res


class StaticListHub(hub.BaseHub):
    def __init__(self, list_module):
        super(StaticListHub, self).__init__()

        collection = pydoc.locate(list_module)

        assert collection is not None, "Module `{list_module}` not found"

        self.list_of_nodes = []
        for func_or_group in collection:
            raw_item = plynx.node.utils.func_or_group_to_dict(func_or_group)
            self.list_of_nodes.append(_enhance_list_item(raw_item))

    def search(self, query):
        # TODO use search_parameters
        # TODO should parse_search_string be removed from nodes_collection?
        _, search_string = parse_search_string(query.search)

        res = _recursive_filter(search_string, self.list_of_nodes)
        return {
            'list': res,
            'metadata': [
                {
                    'total': len(res)
                },
            ],
        }



from dataclasses import dataclass, field
from typing import Any, List, Optional
import cloudpickle
import plynx.node.typing as typings

import plynx.node.utils as utils    # noqa


@dataclass
class InputItem:
    name: str
    file_type: str
    is_array: bool
    min_count: int

    def to_dict(self):
        return {
            "name": self.name,
            "file_type": self.file_type,
            "is_array": self.is_array,
            "min_count": self.min_count,
        }


@dataclass
class OutputItem:
    name: str
    file_type: str
    is_array: bool
    min_count: int

    def to_dict(self):
        return {
            "name": self.name,
            "file_type": self.file_type,
            "is_array": self.is_array,
            "min_count": self.min_count,
        }


@dataclass
class ParamItem:
    name: str
    parameter_type: str
    value: Any
    widget: Optional[str]

    def to_dict(self):
        return {
            "name": self.name,
            "parameter_type": self.parameter_type,
            "value": self.value,
            "widget": self.widget,
        }


@dataclass
class PlynxParams:
    title: str
    description: str
    kind: str
    node_type: str
    inputs: List[InputItem] = field(default_factory=list)
    params: List[ParamItem] = field(default_factory=list)
    outputs: List[OutputItem] = field(default_factory=list)


def input(
    name=None,
    var_type=None,
    is_array=False,
    min_count=1,
):
    def decorator(func_or_class):
        func_or_class.plynx_params.inputs.insert(0, InputItem(
            name=name,
            file_type=typings.type_to_str(var_type),
            is_array=is_array,
            min_count=min_count,
        ))
        return func_or_class
    return decorator


def output(
    name=None,
    var_type=None,
    is_array=False,
    min_count=1,
):
    def decorator(func_or_class):
        func_or_class.plynx_params.outputs.insert(0, OutputItem(
            name=name,
            file_type=typings.type_to_str(var_type),
            is_array=is_array,
            min_count=min_count,
        ))
        return func_or_class
    return decorator


def param(
    name=None,
    var_type=None,
    default=None,
    widget="",
):
    def decorator(func_or_class):
        func_or_class.plynx_params.params.insert(0, ParamItem(
            name=name,
            parameter_type=typings.type_to_str(var_type),
            value=default,
            widget=None if widget is None else widget or name,
        ))
        return func_or_class
    return decorator


def operation(node_type=None, title=None, description="", kind=None):
    def decorator(func_or_class):
        func_or_class.plynx_params = PlynxParams(
            title=title or func_or_class.__name__,
            description=description,
            kind=kind or "python-code-operation",
            node_type=node_type or "python-code-operation",
            inputs=[],
            params=[],
            outputs=[],
        )

        fn_dumped = cloudpickle.dumps(func_or_class).hex()
        func_or_class.plynx_params.params.append(ParamItem(
            name="_pickled_fn",
            parameter_type="str",
            value=fn_dumped,
            widget=None,
        ))
        return func_or_class
    return decorator


parameter = param   # Alias



from subprocess import Popen
import os
import time
import uuid
import shutil
import string
import random
import logging
import collections
import kubernetes
from enum import Enum
from plynx.constants import NodeRunningStatus, ParameterTypes
from plynx.db.node import Parameter, Output
import plynx.plugins.executors.local as local
from plynx.plugins.resources.common import FILE_KIND

NAMESPACE = 'plynx-worker'
SCHEDULING_RETRY = 30

KUBECTL_RETURN_CODES = {
    137: 'Pod was not found: plese check `_timeout` parameter and/or `k8s_worker` logs',
}

KeyValue = collections.namedtuple('KeyValue', ['name', 'default', 'type'])


class KeyConstants(Enum):
    IMAGE = KeyValue('_image', 'alpine:3.7', ParameterTypes.STR)
    IMAGE_COMMAND = KeyValue('_image_command', 'sh', ParameterTypes.STR)
    CPU = KeyValue('_cpu', '100m', ParameterTypes.STR)
    MEMORY = KeyValue('_memory', '1024Mi', ParameterTypes.STR)
    GPU = KeyValue('_gpu', 0, ParameterTypes.INT)
    SCHEDULING_RETRY = KeyValue('_scheduling_retry', 30, ParameterTypes.INT)


def gen_rand(length=8):
    letters = string.ascii_lowercase
    return ''.join(random.choice(letters) for i in range(length))


def gen_rand_name():
    return '{}-{}'.format('job', gen_rand())


def _extend_default_node_in_place(node):
    for param in KeyConstants:
        node.parameters.append(
            Parameter.from_dict({
                'name': param.value.name,
                'parameter_type': param.value.type,
                'value': param.value.default,
                'mutable_type': False,
                'publicable': True,
                'removable': False,
                }
            )
        )

    node.logs.extend(
        [
            Output({
                'name': 'k8s_worker',
                'file_type': FILE_KIND,
            }),
        ]
    )


def _extract_param_value(node_param_dict, e):
    return node_param_dict.get(e.value.name, e.value.default)


class KubernetesStatusPhase:
    PENDING = 'Pending'
    RUNNING = 'Running'
    SUCCEEDED = 'Succeeded'
    FAILED = 'Failed'
    UNKNOWN = 'Unknown'


def get_param_dict(node):
    res = {}
    for parameter in node.parameters:
        res[parameter.name] = parameter.value
    return res


_api_instance = None


def get_api_instance():
    global _api_instance
    if _api_instance:
        return _api_instance
    kubernetes.config.load_kube_config()
    configuration = kubernetes.client.Configuration()
    _api_instance = kubernetes.client.CoreV1Api(kubernetes.client.ApiClient(configuration))
    return _api_instance


def wait_and_check_return_status(sp):
    returncode = sp.wait()

    if returncode == 0:
        return
    if returncode in KUBECTL_RETURN_CODES:
        raise Exception(KUBECTL_RETURN_CODES[returncode])
    raise Exception('Error: unexpected return code `{}`. Please check the logs for details.'.format(returncode))


def create_kubernetes_body(node_param_dict, job_name, workdir):
    container_image = _extract_param_value(node_param_dict, KeyConstants.IMAGE)
    ENV_LIST = []

    body = kubernetes.client.V1Pod(api_version="v1", kind="Pod")
    body.metadata = kubernetes.client.V1ObjectMeta(namespace=NAMESPACE, name=job_name)
    body.status = kubernetes.client.V1PodStatus()

    # Now we start with the Template...
    template = kubernetes.client.V1PodTemplate()
    template.template = kubernetes.client.V1PodTemplateSpec()

    container = kubernetes.client.V1Container(
        name=job_name,
        image=container_image,
        env=ENV_LIST,
        command=['sleep', str(node_param_dict.get('_timeout', 600) * 60)],
        resources=kubernetes.client.V1ResourceRequirements(
            requests={
                'memory': _extract_param_value(node_param_dict, KeyConstants.MEMORY),
                'cpu': _extract_param_value(node_param_dict, KeyConstants.CPU),
                'nvidia.com/gpu': _extract_param_value(node_param_dict, KeyConstants.GPU),
            },
            limits={
                'memory': _extract_param_value(node_param_dict, KeyConstants.MEMORY),
                'cpu': _extract_param_value(node_param_dict, KeyConstants.CPU),
                'nvidia.com/gpu': _extract_param_value(node_param_dict, KeyConstants.GPU),
            },
        ),
    )
    template.template.spec = kubernetes.client.V1PodSpec(
        containers=[container],
        restart_policy='Never',
        )
    # And finaly we can create our V1JobSpec!
    body.spec = kubernetes.client.V1PodSpec(
        containers=[container],
        restart_policy='Never',
        )

    return body


def delete_pod(job_name, log_stream):
    logging.info("Deleting pod")
    sp = Popen(
        [
            'kubectl', 'delete', 'pod',
            '--namespace', NAMESPACE,
            '--grace-period=0',
            job_name
        ],
        stdout=log_stream,
        stderr=log_stream,
    )
    wait_and_check_return_status(sp)


def _init(self):
    self.job_name = None


def _exec_script(self, script_location):
    self._node_running_status = NodeRunningStatus.FAILED

    try:
        param_dict = get_param_dict(self.node)
        self.job_name = gen_rand_name()
        body = create_kubernetes_body(param_dict, self.job_name, self.workdir)
        api_response = get_api_instance().create_namespaced_pod(NAMESPACE, body, pretty=True)
        logging.info(api_response)

        # append running script to worker log
        with open(script_location, 'r') as sf, open(self.logs['worker'], 'a') as wf:
            wf.write(self._make_debug_text("Running script:"))
            wf.write(sf.read())
            wf.write('\n')
            wf.write(self._make_debug_text("End script"))

        stream = kubernetes.watch.Watch().stream(
            func=get_api_instance().list_namespaced_pod,
            namespace=NAMESPACE,
            field_selector="metadata.name={}".format(self.job_name)
            )
        second_time_running = False
        with open(self.logs['stdout'], 'wb', buffering=0) as stdout_file, open(self.logs['stderr'], 'wb', buffering=0) as stderr_file:
            for event in stream:
                phase = event["object"].status.phase
                with open(self.logs['k8s_worker'], 'a+') as k8s_worker_log_file:
                    k8s_worker_log_file.write(
                        'Pod `{}` in `{}` status\n'.format(self.job_name, phase)
                        )
                    if phase == KubernetesStatusPhase.PENDING:
                        # for some reason pod sometimes gets back to this state
                        if second_time_running:
                            break
                        if not event['object'].status.conditions:
                            continue
                        for condition in event['object'].status.conditions:
                            if condition.reason == 'Unschedulable':
                                scheduled = False
                                for ii in range(param_dict.get('_scheduling_retry', SCHEDULING_RETRY)):
                                    time.sleep(1)
                                    pod_list = get_api_instance().list_namespaced_pod(
                                        namespace=NAMESPACE,
                                        field_selector="metadata.name={}".format(self.job_name)
                                    )
                                    k8s_worker_log_file.write(
                                        'Retry scheduling #{}\n'.format(ii)
                                        )
                                    if len(pod_list.items) != 1:
                                        raise Exception('Unexpected number of pods: {}'.format(len(pod_list.items)))
                                    if pod_list.items[0].status.phase == KubernetesStatusPhase.PENDING:
                                        continue
                                    scheduled = True
                                    break
                                if not scheduled:
                                    raise Exception(condition.message)
                    elif phase == KubernetesStatusPhase.RUNNING:
                        if second_time_running:
                            continue
                        second_time_running = True
                        k8s_worker_log_file.write('Uploading artifacts...\n')
                        sp = Popen(
                            [
                                'kubectl', 'cp',
                                '--namespace', NAMESPACE,
                                self.workdir, '{}:{}'.format(self.job_name, self.workdir)
                            ],
                            stdout=k8s_worker_log_file,
                            stderr=k8s_worker_log_file,
                            cwd=self.workdir
                        )
                        wait_and_check_return_status(sp)

                        k8s_worker_log_file.write('Running script...\n')
                        sp = Popen(
                            [
                                'kubectl', 'exec',
                                '--namespace', NAMESPACE,
                                '-it', self.job_name,
                                '--',
                                _extract_param_value(param_dict, KeyConstants.IMAGE_COMMAND), script_location
                            ],
                            stdout=stdout_file,
                            stderr=stderr_file,
                            cwd=self.workdir
                        )
                        wait_and_check_return_status(sp)

                        # Hack: absolute path does not work with `kubectl cp`
                        # solution: download to a local `tmp_dir` directory and move it to `workdir`
                        tmp_dir = str(uuid.uuid1())
                        os.mkdir(tmp_dir)
                        k8s_worker_log_file.write('Downloading artifacts...\n')
                        for output_name, filename in self.output_to_filename.items():
                            k8s_worker_log_file.write('Downloading `{}`...\n'.format(output_name))
                            local_filename = os.path.join(tmp_dir, os.path.basename(filename))
                            sp = Popen(
                                [
                                    'kubectl', 'cp',
                                    '--namespace', NAMESPACE,
                                    '{}:{}'.format(self.job_name, filename), local_filename,
                                ],
                                stdout=k8s_worker_log_file,
                                stderr=k8s_worker_log_file,
                            )
                            wait_and_check_return_status(sp)

                            if os.path.exists(filename):
                                if os.path.isfile(filename):
                                    os.remove(filename)
                                else:
                                    shutil.rmtree(filename)
                            shutil.move(local_filename, filename)
                        os.rmdir(tmp_dir)

                        delete_pod(self.job_name, k8s_worker_log_file)
                        self._node_running_status = NodeRunningStatus.SUCCESS
                    elif phase in (KubernetesStatusPhase.SUCCEEDED, KubernetesStatusPhase.FAILED):
                        if self._node_running_status == NodeRunningStatus.SUCCESS:
                            break
                        k8s_worker_log_file.write('Removing pod `{}`\n'.format(self.job_name))
                        delete_pod(self.job_name, k8s_worker_log_file)
                        raise Exception("Kubernetes pod failed")
                    elif phase == KubernetesStatusPhase.UNKNOWN:
                        raise Exception("Kubernetes pod is unknown")
                    else:
                        raise Exception("Unknown Kubernetes pod phase: `{}`".format(phase))

    except Exception as e:
        self._node_running_status = NodeRunningStatus.FAILED
        logging.exception("Job failed")
        with open(self.logs['k8s_worker'], 'a+') as k8s_worker_log_file:
            k8s_worker_log_file.write(self._make_debug_text("JOB FAILED"))
            k8s_worker_log_file.write(str(e))

    return self._node_running_status


def _kill(self):
    if not self.job_name:
        return

    self._node_running_status = NodeRunningStatus.CANCELED
    delete_pod(self.job_name, None)


class BashJinja2(local.BashJinja2):
    def __init__(self, node=None):
        super(BashJinja2, self).__init__(node)
        _init(self)

    @classmethod
    def get_default_node(cls, is_workflow):
        node = super().get_default_node(is_workflow)
        node.title = 'New bash k8s script'

        _extend_default_node_in_place(node)

        return node

    def exec_script(self, script_location):
        return _exec_script(self, script_location)

    def kill(self):
        return _kill(self)


class PythonNode(local.PythonNode):
    def __init__(self, node=None):
        super(PythonNode, self).__init__(node)
        _init(self)

    @classmethod
    def get_default_node(cls, is_workflow):
        node = super().get_default_node(is_workflow)
        node.title = 'New python k8s script'

        _extend_default_node_in_place(node)

        param = list(filter(lambda p: p.name == '_cmd', node.parameters))[0]
        param.value.mode = 'python'
        param.value.value = 'print("hello world")'

        param = list(filter(lambda p: p.name == KeyConstants.IMAGE_COMMAND.value.name, node.parameters))[0]
        param.value = 'python'

        return node

    def exec_script(self, script_location):
        return _exec_script(self, script_location)

    def kill(self):
        return _kill(self)



import logging


class Registry:
    def __init__(self):
        self.code_function_location_to_node = {}
        self.code_function_location_to_hash = {}

    def register_node(self, node):
        self.code_function_location_to_node[node.code_function_location] = node
        self.code_function_location_to_hash[node.code_function_location] = node.code_hash
        logging.warn(f"Registered operation with location `{node.code_function_location}`")

    def find_nodes(self, function_locations):
        res = []
        for fl in function_locations:
            if fl not in self.code_function_location_to_node:
                continue
            res.append(self.code_function_location_to_node[fl])
        return res


registry = Registry()



import plynx.node


@plynx.node.param(name='your_name', var_type=str, default="World")
@plynx.node.output(name='name', var_type=str)
@plynx.node.operation(
    title="Get name",
)
def get_name(your_name):
    return {
        "name": your_name,
    }


@plynx.node.input(name='name', var_type=str)
@plynx.node.output(name='message', var_type=str)
@plynx.node.operation(
    title="Print message",
)
def print_message(name):
    res = f"Hello {name}!"
    print(res)
    return {
        'message': res,
    }


GROUP = plynx.node.utils.Group(
    title="Hello World",
    items=[
        get_name,
        print_message,
    ]
)



import math
import plynx.node


def print_any(foo):
    print(str(type(foo)).replace("<", "").replace(">", ""))
    print(foo)


@plynx.node.output(name='int', var_type=int)
@plynx.node.output(name='str', var_type=str)
@plynx.node.output(name='dict', var_type=dict)
@plynx.node.output(name='float', var_type=float)
@plynx.node.output(name='bool', var_type=bool)
@plynx.node.operation(
    title="Produce Values",
    description="Echo all types",
)
def all_types():
    return {
        "int": 42,
        "str": "Hello world",
        "dict": {"foo": "woo"},
        "float": math.pi,
        "bool": True,
    }


@plynx.node.input(name='foo', var_type=int)
@plynx.node.operation(
    title="Print Int value",
)
def print_int(foo):
    print_any(foo)


@plynx.node.input(name='foo', var_type=str)
@plynx.node.operation(
    title="Print String value",
)
def print_str(foo):
    print_any(foo)


@plynx.node.input(name='foo', var_type=dict)
@plynx.node.operation(
    title="Print Dict value",
)
def print_dict(foo):
    print_any(foo)


@plynx.node.input(name='foo', var_type=float)
@plynx.node.operation(
    title="Print Float value",
)
def print_float(foo):
    print_any(foo)


@plynx.node.input(name='foo', var_type=bool)
@plynx.node.operation(
    title="Print Bool value",
)
def print_bool(foo):
    print_any(foo)


GROUP = plynx.node.utils.Group(
    title="Test Python built-in types",
    items=[
        all_types,
        print_int,
        print_str,
        print_dict,
        print_float,
        print_bool,
    ]
)



ANY = "any"

STR_TO_CLASS = {
    "int": int,
    "str": str,
    "dict": dict,
    "float": float,
    "bool": bool,
    # Any other
    ANY: lambda x: x,
}

CLASS_TO_STR = {
    v: k for k, v in STR_TO_CLASS.items()
}


def type_to_str(tp):
    return CLASS_TO_STR[tp]



from flask import g, request
import json
import plynx.db.node_collection_manager
from plynx.db.db_object import get_class
from plynx.db.demo_user_manager import DemoUserManager
from plynx.db.user import UserCollectionManager
from plynx.web.common import app, register_user, requires_auth, make_success_response, make_fail_response, handle_errors
from plynx.utils.common import JSONEncoder, to_object_id
from plynx.utils.exceptions import RegisterUserException
from plynx.constants import Collections, NodeClonePolicy, IAMPolicies, UserPostAction
from plynx.db.user import User

demo_user_manager = DemoUserManager()
template_collection_manager = plynx.db.node_collection_manager.NodeCollectionManager(collection=Collections.TEMPLATES)


@app.route('/plynx/api/v0/token', strict_slashes=False)
@requires_auth
@handle_errors
def get_auth_token():
    access_token = g.user.generate_access_token()
    refresh_token = g.user.generate_refresh_token()

    user_obj = g.user.to_dict()
    user_obj['hash_password'] = ''
    return make_success_response({
        'access_token': access_token.decode('ascii'),
        'refresh_token': refresh_token.decode('ascii'),
        'user': user_obj,
    })


@app.route('/plynx/api/v0/demo', methods=['POST'])
@handle_errors
def post_demo_user():
    user = demo_user_manager.create_demo_user()
    if not user:
        return make_fail_response('Failed to create a demo user')

    template_id = DemoUserManager.demo_config.kind
    if DemoUserManager.demo_config.template_id:
        try:
            node_id = to_object_id(DemoUserManager.demo_config.template_id)
        except Exception as e:
            app.logger.error('node_id `{}` is invalid'.format(DemoUserManager.demo_config.template_id))
            app.logger.error(e)
            return make_fail_response('Failed to create a demo node')
        try:
            user_id = user._id
            node = template_collection_manager.get_db_node(node_id, user_id)
            node = get_class(node['_type'])(node).clone(NodeClonePolicy.NODE_TO_NODE)
            node.author = user_id
            node.save()
            template_id = node._id
        except Exception as e:
            app.logger.error('Failed to create a demo node')
            app.logger.error(e)
            return make_fail_response(str(e)), 500

    access_token = user.generate_access_token(expiration=1800)
    user_obj = user.to_dict()
    user_obj['hash_password'] = ''
    return JSONEncoder().encode({
        'access_token': access_token.decode('ascii'),
        'refresh_token': 'Not assigned',
        'user': user_obj,
        'url': '/{}/{}'.format(Collections.TEMPLATES, template_id),
    })


@app.route('/plynx/api/v0/users/<username>', methods=['GET'])
@handle_errors
@requires_auth
def get_user(username):
    user = UserCollectionManager.find_user_by_name(username)
    if not user:
        return make_fail_response('User not found'), 404
    user_obj = user.to_dict()

    is_admin = IAMPolicies.IS_ADMIN in g.user.policies
    user_obj['_is_admin'] = is_admin
    user_obj['_readonly'] = user._id != g.user._id and not is_admin
    del user_obj['password_hash']

    return make_success_response({
        'user': user_obj,
        })


@app.route('/plynx/api/v0/users', methods=['POST'])
@handle_errors
@requires_auth
def post_user():
    data = json.loads(request.data)
    app.logger.warn(data)
    action = data.get('action', '')
    old_password = data.get('old_password', '')
    new_password = data.get('new_password', '')
    if action == UserPostAction.MODIFY:
        posted_user = User.from_dict(data['user'])
        existing_user = UserCollectionManager.find_user_by_name(posted_user.username)
        if not existing_user:
            return make_fail_response('User not found'), 404
        if g.user.username != posted_user.username and IAMPolicies.IS_ADMIN not in g.user.policies:
            return make_fail_response('You don`t have permission to modify this user'), 401

        if set(posted_user.policies) != set(existing_user.policies):
            if IAMPolicies.IS_ADMIN not in g.user.policies:
                return make_fail_response('You don`t have permission to modify policies'), 401
            existing_user.policies = posted_user.policies

        if new_password:
            if not existing_user.verify_password(old_password):
                return make_fail_response('Incorrect password'), 401
            existing_user.hash_password(new_password)

        existing_user.settings = posted_user.settings

        existing_user.save()
        if g.user.username == posted_user.username:
            g.user = posted_user

        is_admin = IAMPolicies.IS_ADMIN in g.user.policies
        user_obj = existing_user.to_dict()
        user_obj['_is_admin'] = is_admin
        user_obj['_readonly'] = existing_user._id != g.user._id and not is_admin
        del user_obj['password_hash']

        return make_success_response({
            'user': user_obj,
            })
    else:
        raise Exception('Unknown action: `{}`'.format(action))

    raise NotImplementedError("Nothing is to return")


@app.route('/plynx/api/v0/register', methods=['POST'])
@handle_errors
def post_register():
    query = json.loads(request.data)

    email = query['email'].lower()
    username = query['username']
    password = query['password']

    try:
        user = register_user(
            username=username,
            password=password,
            email=email,
        )
    except RegisterUserException as ex:
        return make_fail_response(
            ex.message,
            error_code=ex.error_code,
        ), 400

    g.user = user
    access_token = user.generate_access_token()
    refresh_token = user.generate_refresh_token()

    user_obj = user.to_dict()
    user_obj['hash_password'] = ''
    return make_success_response({
        'access_token': access_token.decode('ascii'),
        'refresh_token': refresh_token.decode('ascii'),
        'user': user_obj,
    })



import logging
import queue
import cloudpickle
import multiprocessing
import traceback
import uuid
import inspect
import sys
import os
import contextlib
import threading
from plynx.constants import NodeRunningStatus, Collections
from plynx.utils.common import to_object_id
import plynx.utils.remote
import plynx.plugins.executors.dag

stateful_init_mutex = threading.Lock()
stateful_class_registry = {}

POOL_SIZE = 3


def materialize_fn(fn_str):
    fn_bytes = bytes.fromhex(fn_str)
    fn = cloudpickle.loads(fn_bytes)
    return fn


def prep_args(node):
    args = {}
    for input in node.inputs:
        args[input.name] = input.values if input.is_array else input.values[0]

    # TODO smater way to determine what parameters to pass
    visible_parameters = filter(lambda param: param.widget is not None, node.parameters)
    args.update(
        plynx.plugins.executors.local.prepare_parameters_for_python(visible_parameters)
    )
    return args


def assign_outputs(node, output_dict):
    if not output_dict:
        return
    for key, value in output_dict.items():
        node_output = node.get_output_by_name(key)
        node_output.values = value if node_output.is_array else [value]


class redirect_to_plynx_logs:
    def __init__(self, node, stdout, stderr):
        self.stdout_filename = str(uuid.uuid4())
        self.stderr_filename = str(uuid.uuid4())

        self.stdout = stdout
        self.stderr = stderr

        self.node = node

        self.names_map = [
            (self.stdout_filename, self.stdout),
            (self.stderr_filename, self.stderr),
        ]

    def __enter__(self):
        self.stdout_file = open(self.stdout_filename, 'w')
        self.stderr_file = open(self.stderr_filename, 'w')

        self.stdout_redirect = contextlib.redirect_stdout(self.stdout_file)
        self.stderr_redirect = contextlib.redirect_stderr(self.stderr_file)

        self.stdout_redirect.__enter__()
        self.stderr_redirect.__enter__()
        return None

    def __exit__(self, *args):

        self.stdout_redirect.__exit__(*sys.exc_info())
        self.stderr_redirect.__exit__(*sys.exc_info())

        self.stdout_file.close()
        self.stderr_file.close()

        for filename, logs_name in self.names_map:
            if os.stat(filename).st_size > 0:
                with plynx.utils.remote.open(filename, "w") as f:
                    with open(filename) as fi:
                        f.write(fi.read())
                output = self.node.get_log_by_name(name=logs_name)
                output.values = [filename]


def worker_main(job_run_queue, job_complete_queue):
    logging.warn("Created pool worker")
    while True:
        node = job_run_queue.get()

        try:
            pickled_fn_parameter = node.get_parameter_by_name("_pickled_fn")

            fn = materialize_fn(pickled_fn_parameter.value)
            if inspect.isclass(fn):
                with stateful_init_mutex:
                    if node.code_hash not in stateful_class_registry:
                        with redirect_to_plynx_logs(node, "init_stdout", "init_stderr"):
                            stateful_class_registry[node.code_hash] = fn()
                    fn = stateful_class_registry[node.code_hash]

            with redirect_to_plynx_logs(node, "stdout", "stderr"):
                res = fn(**prep_args(node))

        except Exception:
            error_str = traceback.format_exc()
            logging.error(f"Job failed with traceback: {error_str}")
            node.node_running_status = NodeRunningStatus.FAILED

            err_filename = str(uuid.uuid4())
            with plynx.utils.remote.open(err_filename, "w") as f:
                f.write(error_str)
            output = node.get_log_by_name(name="traceback")
            output.values = [err_filename]
            logging.error(f"Wrote to logs: {err_filename}")
        else:
            assign_outputs(node, res)
            node.node_running_status = NodeRunningStatus.SUCCESS
        job_complete_queue.put(node)

    logging.warn("Deleted pool worker")


class DAG(plynx.plugins.executors.dag.DAG):
    """ Python DAG scheduler.

    Args:
        node_dict (dict)
    """
    IS_GRAPH = True
    GRAPH_ITERATION_SLEEP = 0

    def __init__(self, node_dict):
        super(DAG, self).__init__(node_dict)

        self.job_run_queue = queue.Queue()
        self.job_complete_queue = queue.Queue()

        self.worker_pool = multiprocessing.pool.ThreadPool(
            POOL_SIZE, worker_main, (
                self.job_run_queue,
                self.job_complete_queue,
            )
        )

    def pop_jobs(self):
        """Get a set of nodes with satisfied dependencies"""
        res = []

        num_completed_jobs = 0
        while not self.job_complete_queue.empty():
            try:
                node = self.job_complete_queue.get_nowait()
                self.update_node(node)
                num_completed_jobs += 1
            except queue.Empty:
                logging.warning("Queue is empty")
                break

        if NodeRunningStatus.is_failed(self._node_running_status):
            if self._node_running_status != NodeRunningStatus.FAILED_WAITING:
                logging.info(f"Job in DAG failed with status {self._node_running_status}, pop_jobs will return []")
            return res

        for node_id in self.dependency_index_to_node_ids[0]:
            """Get the node and init its inputs, i.e. filling its resource_ids"""
            node = self.node_id_to_node[node_id]
            for node_input in node.inputs:
                for input_reference in node_input.input_references:
                    node_input.values.extend(
                        self.node_id_to_node[to_object_id(input_reference.node_id)].get_output_by_name(
                            input_reference.output_id
                        ).values
                    )
            node.node_running_status = NodeRunningStatus.IN_QUEUE

            res.append(node)
        del self.dependency_index_to_node_ids[0]

        if num_completed_jobs > 0:
            self.node.save(collection=Collections.RUNS, force=True)

        return res

    def _execute_node(self, node):
        if NodeRunningStatus.is_finished(node.node_running_status):     # NodeRunningStatus.SPECIAL
            return

        logging.info(f'Execute {node} {node.title}')
        node.node_running_status = NodeRunningStatus.RUNNING
        self.node.save(collection=Collections.RUNS)
        # TODO somehow optimize `update_node`?
        # If not copy but original sent, the dependencies list won't be updated
        self.job_run_queue.put(node.copy())

    def kill(self):
        """Force to kill the process.

        The reason can be the fact it was working too long or parent exectuter canceled it.
        """
        logging.warn("Received kill request")
        self._node_running_status = NodeRunningStatus.CANCELED
        self.worker_pool.terminate()



import inspect
import hashlib

from dataclasses import dataclass
from plynx.constants import NodeStatus, NodeRunningStatus, NodeOrigin
from plynx.utils.common import ObjectId


@dataclass
class VersionData:
    unique_name: str
    hash_value: int


@dataclass
class Group:
    title: str
    items: list

    def to_dict(self):
        return {
            "_id": str(ObjectId()),
            "_type": "Group",
            "title": self.title,
            "items": list(map(func_or_group_to_dict, self.items)),
        }


def callable_to_version_data(callable):
    hash_value = hashlib.md5(inspect.getsource(callable).encode()).hexdigest()
    filename = inspect.getfile(callable)
    return VersionData(
        unique_name=f"{filename}:{callable.__name__}",
        hash_value=hash_value,
    )


def func_or_group_to_dict(func_or_group):
    if isinstance(func_or_group, Group):
        return func_or_group.to_dict()
    plynx_params = func_or_group.plynx_params

    version_data = callable_to_version_data(func_or_group)
    return {
        "_id": str(ObjectId()),
        "_type": "Node",
        "title": plynx_params.title,
        "description": plynx_params.description,
        "node_running_status": NodeRunningStatus.CREATED,
        "node_status": NodeStatus.READY,
        "author": None,
        "kind": plynx_params.kind,
        "origin": NodeOrigin.BUILT_IN_HUBS,
        "code_hash": version_data.hash_value,
        "code_function_location": version_data.unique_name,
        "inputs": list(map(lambda item: item.to_dict(), plynx_params.inputs)),
        "outputs": list(map(lambda item: item.to_dict(), plynx_params.outputs)),
        "parameters": list(map(lambda item: item.to_dict(), plynx_params.params)),
        "logs": [],
    }



import math
import time
import random
import sys

import plynx.node


@plynx.node.param(name='value', var_type=int, default=1)
@plynx.node.output(name='x', var_type=int)
@plynx.node.operation(
    title="Make integer"
)
def make_int(value):
    return {
        "x": value,
    }


@plynx.node.input(name='x', var_type=int)
@plynx.node.input(name='y', var_type=int)
@plynx.node.output(name='res', var_type=str)
@plynx.node.param(name='coef', var_type=int, default=1)
@plynx.node.operation(
    title="Multiple inputs",
    description="x * coef + y * pi",
)
def example_func(x, y, coef):
    print(f"x: {type(x)}")
    print(f"y: {type(y)}")
    print(f"coef {type(coef)}")
    res = x * coef + y * math.pi
    return {'res': str(res)}


@plynx.node.input(name='x', var_type=int, is_array=True, min_count=0)
@plynx.node.param(name='timeout', var_type=int, default=3)
@plynx.node.output(name='x', var_type=int)
@plynx.node.operation(
    title="Sleep",
    description="Sleep for timeout sec and add 1.",
)
def sleep(x, timeout):
    print(f"Sleeping for {timeout} sec")
    time.sleep(timeout)
    return {
        "x": sum(x) + 1
    }


@plynx.node.input(name='x', var_type=int, is_array=True, min_count=0)
@plynx.node.output(name='x', var_type=int)
@plynx.node.operation(
    title="Raise exception",
    description="Always raise exception",
)
def error(x):
    raise Exception("Expected exception")
    return {
        "x": sum(x) + 1
    }


@plynx.node.input(name='x', var_type=int, is_array=True, min_count=0)
@plynx.node.output(name='x', var_type=int)
@plynx.node.operation(
    title="Stateful sum",
    description="Add 1 and keep the previous value",
)
class Statefull:
    def __init__(self):
        self.accum = 0
        self.random = random.random()
        print(f"Init worker with id {self.random}")

    def __call__(self, x):
        self.accum += sum(x) + 1
        print(f"I am {self.random} living on {id(self)}")
        print(f"x: {x}")
        print(f"self.accum: {self.accum}")
        print("I am error", file=sys.stderr)
        return {
            "x": self.accum
        }


GROUP = plynx.node.utils.Group(
    title="Test basics",
    items=[
        make_int,
        example_func,
        sleep,
        error,
        Statefull,
    ]
)



from plynx.demo.hello_world import GROUP as hello_group
from plynx.demo.types import GROUP as types_group
from plynx.demo.basic_functions import GROUP as basic_group


COLLECTION = [
    hello_group,
    types_group,
    basic_group,
]
